/*  seek.c - functions related to real-time seeking
 *  Copyright (C) 2000-2001  Jason Jordan (shnutils@freeshell.org)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <xmms/util.h>
#include <xmms/configfile.h>
#include "shn.h"
#include "shorten.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

shn_seek_entry *seek_entry_search(shn_seek_entry *table,long goal,long min,long max)
{
	long med = (min + max) / 2;
	shn_seek_entry *middle = table + med;

	if (goal < middle->shnSample)
		return seek_entry_search(table,goal,min,med-1);
	if (goal > middle->shnSample + 25600)
		return seek_entry_search(table,goal,med+1,max);
	return middle;
}

int load_separate_seek_table(char *filename,shn_file *this_shn)
{
	FILE *f;
	long seek_table_len;

	debug("Looking for seek table in separate file: '%s'",filename);

	if (!(f=fopen(filename,"r")))
	{
		return 0;
	}

	fseek(f,0,SEEK_END);
	seek_table_len = (long)ftell(f) - sizeof(shn_seek_header);
	fseek(f,0,SEEK_SET);

	if (fread((void *)&this_shn->seek_header,1,sizeof(shn_seek_header),f) == sizeof(shn_seek_header))
	{
		this_shn->seek_header.version = uchar_to_ulong_le((unsigned char *)&this_shn->seek_header.version);
		this_shn->seek_header.shnFileSize = uchar_to_ulong_le((unsigned char *)&this_shn->seek_header.shnFileSize);
		if (!memcmp(this_shn->seek_header.signature,SEEK_HEADER_SIGNATURE,sizeof(this_shn->seek_header.signature)))
		{
			if (this_shn->seek_header.shnFileSize == this_shn->wave_header.actual_size)
			{
				if ((this_shn->seek_table = malloc(seek_table_len)))
				{
					if (fread((void *)this_shn->seek_table,1,seek_table_len,f) == seek_table_len)
					{
						debug("Successfully loaded seek table in separate file: '%s'",filename);
						this_shn->vars.seek_table_entries = seek_table_len / sizeof(shn_seek_entry);
						fclose(f);
						convert_seek_tables_to_le(this_shn);
						return 1;
					}
				}
			}
			else
			{
				debug("Seek table .shn file size and actual .shn file size differ");
			}
		}
	}

	fclose(f);
	return 0;
}

void load_seek_table(shn_file *this_shn,char *filename)
{
	char *altfilename,*p;
	int len;

	debug("Looking for seek table appended to file: '%s'",filename);

	fseek(this_shn->vars.fd,-sizeof(shn_seek_trailer),SEEK_END);
	if (fread((void *)&this_shn->seek_trailer,1,sizeof(shn_seek_trailer),this_shn->vars.fd) == sizeof(shn_seek_trailer))
	{
		this_shn->seek_trailer.seekTableSize = uchar_to_ulong_le((unsigned char *)&this_shn->seek_trailer.seekTableSize);
		if (!memcmp(this_shn->seek_trailer.signature,SEEK_TRAILER_SIGNATURE,sizeof(this_shn->seek_trailer.signature)))
		{
			fseek(this_shn->vars.fd,-this_shn->seek_trailer.seekTableSize,SEEK_END);
			this_shn->seek_trailer.seekTableSize -= (sizeof(shn_seek_header) + sizeof(shn_seek_trailer));
			if (fread((void *)&this_shn->seek_header,1,sizeof(shn_seek_header),this_shn->vars.fd) == sizeof(shn_seek_header))
			{
				if ((this_shn->seek_table = malloc(this_shn->seek_trailer.seekTableSize)))
				{
					if (fread((void *)this_shn->seek_table,1,this_shn->seek_trailer.seekTableSize,this_shn->vars.fd) == this_shn->seek_trailer.seekTableSize)
					{
						debug("Successfully loaded seek table appended to file: '%s'",filename);
						this_shn->vars.seek_table_entries = this_shn->seek_trailer.seekTableSize / sizeof(shn_seek_entry);
						convert_seek_tables_to_le(this_shn);
						return;
					}
				}
			}
		}
	}

	/* otherwise, look for it in a separate file */

	if (!(altfilename = malloc(strlen(filename)+sizeof(SEEK_SUFFIX)+1)))
	{
		debug("Could not allocate memory for alternate filename");
		return;
	}

	strcpy(altfilename,filename);
	if (filename_contains_a_dot(altfilename))
		p = strrchr(altfilename,'.');
	else
		p = altfilename + strlen(altfilename);
	strcpy(p,SEEK_SUFFIX);

	if (load_separate_seek_table(altfilename,this_shn))
	{
		free(altfilename);
		return;
	}

	free(altfilename);

	if ((p=strrchr(filename,'/')))
	{
		p++;
		len = filename + strlen(filename) - p + 1;
	}
	else
	{
		p = filename;
		len = strlen(filename);
	}

	if (!(altfilename = malloc(strlen(shn_cfg.seek_tables_path)+1+len+sizeof(SEEK_SUFFIX)+1)))
	{
		debug("Could not allocate memory for alternate filename");
		return;
	}

	strcpy(altfilename,shn_cfg.seek_tables_path);
	strcat(altfilename,"/");
	strcat(altfilename,p);
	if (filename_contains_a_dot(altfilename))
		p = strrchr(altfilename,'.');
	else
		p = altfilename + strlen(altfilename);
	strcpy(p,SEEK_SUFFIX);

	if (load_separate_seek_table(altfilename,this_shn))
	{
		free(altfilename);
		return;
	}

	debug("Could not find any seek tables");
	free(altfilename);
}

void convert_seek_tables_to_le(shn_file *this_shn)
{
	shn_seek_entry *seek_info = this_shn->seek_table;
	int i;

	for(i=0;i<this_shn->vars.seek_table_entries;i++)
	{
		seek_info->shnSample       = uchar_to_ulong_le((unsigned char *)&seek_info->shnSample);
		seek_info->shnByteOffset   = uchar_to_ulong_le((unsigned char *)&seek_info->shnByteOffset);
		seek_info->shnLastPosition = uchar_to_ulong_le((unsigned char *)&seek_info->shnLastPosition);
		seek_info->shnByteGet      = uchar_to_ushort_le((unsigned char *)&seek_info->shnByteGet);
		seek_info->shnBufferOffset = uchar_to_ushort_le((unsigned char *)&seek_info->shnBufferOffset);
		seek_info->shnBitOffset    = uchar_to_ushort_le((unsigned char *)&seek_info->shnBitOffset);
		seek_info->shnGBuffer      = uchar_to_ulong_le((unsigned char *)&seek_info->shnGBuffer);
		seek_info->shnBitShift     = uchar_to_ushort_le((unsigned char *)&seek_info->shnBitShift);
		seek_info->cbuf0[0]        = uchar_to_long_le((unsigned char *)&seek_info->cbuf0[0]);
		seek_info->cbuf0[1]        = uchar_to_long_le((unsigned char *)&seek_info->cbuf0[1]);
		seek_info->cbuf0[2]        = uchar_to_long_le((unsigned char *)&seek_info->cbuf0[2]);
		seek_info->cbuf1[0]        = uchar_to_long_le((unsigned char *)&seek_info->cbuf1[0]);
		seek_info->cbuf1[1]        = uchar_to_long_le((unsigned char *)&seek_info->cbuf1[1]);
		seek_info->cbuf1[2]        = uchar_to_long_le((unsigned char *)&seek_info->cbuf1[2]);
		seek_info->offset0[0]      = uchar_to_long_le((unsigned char *)&seek_info->offset0[0]);
		seek_info->offset0[1]      = uchar_to_long_le((unsigned char *)&seek_info->offset0[1]);
		seek_info->offset0[2]      = uchar_to_long_le((unsigned char *)&seek_info->offset0[2]);
		seek_info->offset0[3]      = uchar_to_long_le((unsigned char *)&seek_info->offset0[3]);
		seek_info->offset1[0]      = uchar_to_long_le((unsigned char *)&seek_info->offset1[0]);
		seek_info->offset1[1]      = uchar_to_long_le((unsigned char *)&seek_info->offset1[1]);
		seek_info->offset1[2]      = uchar_to_long_le((unsigned char *)&seek_info->offset1[2]);
		seek_info->offset1[3]      = uchar_to_long_le((unsigned char *)&seek_info->offset1[3]);
		seek_info++;
	}
}
